/* 
 * Copyright (C) 2020 wea_ondara
 *
 * BungeePerms is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BungeePerms is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.alpenblock.bungeeperms.config;

import java.io.File;
import java.io.InputStream;
import java.util.Map;

import org.yaml.snakeyaml.DumperOptions;
import org.yaml.snakeyaml.Yaml;
import org.yaml.snakeyaml.error.YAMLException;
import org.yaml.snakeyaml.representer.Representer;

/**
 * An implementation of {@link Configuration} which saves all files in Yaml. Note that this implementation is not synchronized.
 */
public class YamlConfiguration extends FileConfiguration
{

    protected static final String COMMENT_PREFIX = "# ";
    protected static final String BLANK_CONFIG = "{}\n";
    private final DumperOptions yamlOptions = new DumperOptions();
    private final Representer yamlRepresenter = new YamlRepresenter();
    private final Yaml yaml = new Yaml(new YamlConstructor(), yamlRepresenter, yamlOptions);

    @Override
    public String saveToString()
    {
        yamlOptions.setIndent(options().indent());
        yamlOptions.setDefaultFlowStyle(DumperOptions.FlowStyle.BLOCK);
        yamlRepresenter.setDefaultFlowStyle(DumperOptions.FlowStyle.BLOCK);

        String header = buildHeader();
        String dump = yaml.dump(getValues(false));

        if (dump.equals(BLANK_CONFIG))
        {
            dump = "";
        }

        return header + dump;
    }

    @Override
    public void loadFromString(String contents) throws InvalidConfigurationException
    {

        Map<?, ?> input;
        try
        {
            input = (Map<?, ?>) yaml.load(contents);
        }
        catch (YAMLException e)
        {
            throw new InvalidConfigurationException(e);
        }
        catch (ClassCastException e)
        {
            throw new InvalidConfigurationException("Top level is not a Map.");
        }

        String header = parseHeader(contents);
        if (header.length() > 0)
        {
            options().header(header);
        }

        if (input != null)
        {
            convertMapsToSections(input, this);
        }
    }

    protected void convertMapsToSections(Map<?, ?> input, ConfigurationSection section)
    {
        for (Map.Entry<?, ?> entry : input.entrySet())
        {
            String key = entry.getKey().toString();
            Object value = entry.getValue();

            if (value instanceof Map)
            {
                convertMapsToSections((Map<?, ?>) value, section.createSection(key));
            }
            else
            {
                section.set(key, value);
            }
        }
    }

    protected String parseHeader(String input)
    {
        String[] lines = input.split("\r?\n", -1);
        StringBuilder result = new StringBuilder();
        boolean readingHeader = true;
        boolean foundHeader = false;

        for (int i = 0; (i < lines.length) && (readingHeader); i++)
        {
            String line = lines[i];

            if (line.startsWith(COMMENT_PREFIX))
            {
                if (i > 0)
                {
                    result.append("\n");
                }

                if (line.length() > COMMENT_PREFIX.length())
                {
                    result.append(line.substring(COMMENT_PREFIX.length()));
                }

                foundHeader = true;
            }
            else if ((foundHeader) && (line.length() == 0))
            {
                result.append("\n");
            }
            else if (foundHeader)
            {
                readingHeader = false;
            }
        }

        return result.toString();
    }

    @Override
    protected String buildHeader()
    {
        String header = options().header();

        if (options().copyHeader())
        {
            Configuration def = getDefaults();

            if ((def != null) && (def instanceof FileConfiguration))
            {
                FileConfiguration filedefaults = (FileConfiguration) def;
                String defaultsHeader = filedefaults.buildHeader();

                if ((defaultsHeader != null) && (defaultsHeader.length() > 0))
                {
                    return defaultsHeader;
                }
            }
        }

        if (header == null)
        {
            return "";
        }

        StringBuilder builder = new StringBuilder();
        String[] lines = header.split("\r?\n", -1);
        boolean startedHeader = false;

        for (int i = lines.length - 1; i >= 0; i--)
        {
            builder.insert(0, "\n");

            if ((startedHeader) || (lines[i].length() != 0))
            {
                builder.insert(0, lines[i]);
                builder.insert(0, COMMENT_PREFIX);
                startedHeader = true;
            }
        }

        return builder.toString();
    }

    @Override
    public YamlConfigurationOptions options()
    {
        if (options == null)
        {
            options = new YamlConfigurationOptions(this);
        }

        return (YamlConfigurationOptions) options;
    }

    public static YamlConfiguration loadConfiguration(File file)
    {

        YamlConfiguration config = new YamlConfiguration();

        try
        {
            config.load(file);
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }
        return config;
    }

    public static YamlConfiguration loadConfiguration(InputStream stream)
    {
        YamlConfiguration config = new YamlConfiguration();

        try
        {
            config.load(stream);
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }
        return config;
    }
}
